MODULE bigfun
! BIG dimensional FUNction box adapted for single shooting
! ----------------------------------------------------------------------
! Version 1.01, last modified 27-01-1999.
! --- Function ---
! Big dimensional (n times n) right hand side used for simultaneous
! integration of the fundamental solution matrix (linear case) or the
! Jacobian of Newton's method (non-linear case) in shooting procedure.
! The ODE is related to the original right hand side of FUNBOX.
! --- Author ---
! Peter W. KOFLER,
! Department of Applied Mathematics and Numerical Analysis,
! Vienna University of Technology, Vienna, Austria.
! ----------------------------------------------------------------------
   USE types                 ! predefined data-types
   USE funbox, ONLY : box_neqn, & ! import dimension,
                      box_fcn, box_dfcn ! ODE and Jacobian
   IMPLICIT NONE
   INTEGER(KIND = i4b), PARAMETER :: neqn = box_neqn
!        neqn ... dimension of the original problem
   INTEGER(KIND = i4b) :: dimp
!        dimp ... number of not separated BCs (<= neqn)
   PRIVATE
   PUBLIC :: big_init, &
             bigl_fcn, bigl_dfcn, bign_fcn, bign_dfcn, &
             dummy1, dummy2
!        SUBROUTINE big_init ...... initialise variables
!        REAL FUNCTION bigl_fcn .... linear f(x,y) (right hand side)
!        REAL FUNCTION bigl_dfcn ... linear df/dy(x,y) (Jacobian)
!        REAL FUNCTION bign_fcn .... non-linear f(x,y) (right hand side)
!        REAL FUNCTION bign_dfcn ... non-linear df/dy(x,y) (Jacobian)
!        REAL FUNCTION dummy1 ...... dummy for box_true
!        REAL FUNCTION dummy2 ...... dummy for box_truelocal
!                 as F90 does not allow OPTIONAL PROCEDURE Parameters

   CONTAINS
! ----------------------------------------------------------------------
SUBROUTINE big_init(q)
!        initialise module variables
   IMPLICIT NONE
   INTEGER(KIND = i4b), INTENT(in) :: q
!        q ... number of not separated BCs (<= neqn)
   dimp = q                  ! set up global used dimension
   RETURN
END SUBROUTINE big_init
! ----------------------------------------------------------------------
FUNCTION bigl_fcn(x, y)
!        big right hand side for Phi'(x) = A(x) Phi(x) for linear case
!        with original ODE as last line (dimp+1)-dimensional in n
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: x, y(:)
!        x ... independent variable
!        y ... value of Phi(x) (dependent variable)
   REAL(KIND = dp) :: bigl_fcn(SIZE(y)), Jac(neqn,neqn), w(neqn)
!        Jac ... Jacobian, which is independent of y
!        w ... workspace vector
   INTEGER(KIND = i4b) :: i
!        i ... counter

   IF (x == 0.0_dp) THEN     ! we are not able to calculate the value of
      bigl_fcn = 0.0_dp      ! Phi(0) automatically.
   ELSE
      w = 0.0_dp             ! just a dummy parameter
      Jac = box_dfcn(x,w)    ! get the linear term for dimp vectors of
      DO i = 1, dimp         ! the fundamental solution matrix
        bigl_fcn((i-1)*neqn+1:i*neqn)=MATMUL(Jac,y((i-1)*neqn+1:i*neqn))
      END DO ! DO i = 1, dimp
      bigl_fcn(dimp*neqn+1:(dimp+1)*neqn) = & ! the special solution
                                 box_fcn(x,y(dimp*neqn+1:(dimp+1)*neqn))
   END IF
   RETURN
END FUNCTION bigl_fcn
! ----------------------------------------------------------------------
FUNCTION bigl_dfcn(x, y)
!        big Jacobian of ODE Phi'(x)=A(x) Phi(x) for linear case = A(x)
!        with original Jacobian as last block
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: x, y(:)
!        x ... independent variable
!        y ... dummy value of Phi(x) because linear case
   REAL(KIND = dp) :: bigl_dfcn(SIZE(y),SIZE(y)),Jac(neqn,neqn),w(neqn)
!        Jac ... Jacobian, which is independent of y
!        w ... workspace vector
   INTEGER(KIND = i4b) :: i
!        i ... counter

   bigl_dfcn = 0.0_dp
   w = 0.0_dp                ! just a dummy parameter
   Jac = box_dfcn(x,w)       ! get the linear term
   DO i = 1, dimp+1
      bigl_dfcn((i-1)*neqn+1:i*neqn,(i-1)*neqn+1:i*neqn) = Jac
   END DO ! DO i = 1, dimp+1
   RETURN
END FUNCTION bigl_dfcn
! ----------------------------------------------------------------------
FUNCTION bign_fcn(x, y)
!        big right hand side in non-linear case to calculate Jacobian by
!        differential quotients for Newton's method of the shooting
!        procedure.
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: x, y(:)
!        x ... independent variable
!        y ... values of y_1(x) to y_q(x) (dependent variables)
   REAL(KIND = dp) :: bign_fcn(SIZE(y))
   INTEGER(KIND = i4b) :: i
!        i ... counter

   DO i = 1, dimp
      bign_fcn((i-1)*neqn+1:i*neqn) = box_fcn(x,y((i-1)*neqn+1:i*neqn))
   END DO ! DO i = 1, dimp
   RETURN
END FUNCTION bign_fcn
! ----------------------------------------------------------------------
FUNCTION bign_dfcn(x, y)
!        big Jacobian of non-linear ODE to calculate the Jacobian by
!        differential quotients for Newton's method of the shooting
!        procedure.
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: x, y(:)
!        x ... independent variable
!        y ... values of y_1(x) to y_q(x) (dependent variables)
   REAL(KIND = dp) :: bign_dfcn(SIZE(y),SIZE(y))
   INTEGER(KIND = i4b) :: i
!        i ... counter

   bign_dfcn = 0.0_dp
   DO i = 1, dimp
      bign_dfcn((i-1)*neqn+1:i*neqn, (i-1)*neqn+1:i*neqn) = &
                                 box_dfcn(x,y((i-1)*neqn+1:i*neqn))
   END DO ! DO i = 1, dimp
   RETURN
END FUNCTION bign_dfcn
! ----------------------------------------------------------------------
FUNCTION dummy1(x,y)
!        dummy vector valued function with 1 argument
   IMPLICIT NONE
   REAL (KIND = dp), INTENT(in) :: x, y(:)
   REAL (KIND = dp) :: dummy1(SIZE(y))
   dummy1 = x
   dummy1 = 0.0_dp
   RETURN
END FUNCTION dummy1
! ----------------------------------------------------------------------
FUNCTION dummy2(x1,x2,y)
!        dummy vector valued function with 2 arguments
   IMPLICIT NONE
   REAL (KIND = dp), INTENT(in) :: x1, x2, y(:)
   REAL (KIND = dp) :: dummy2(SIZE(y))
   dummy2 = x1+x2
   dummy2 = 0.0_dp
   RETURN
END FUNCTION dummy2

END MODULE bigfun
