MODULE types
! Define data types
   IMPLICIT NONE
   INTEGER, PARAMETER :: dp = KIND(1.0D0), & ! working precision real
                         i4b = SELECTED_INT_KIND(9)
END MODULE types

MODULE linalg
! --- Function and Algorithm ---
! This module provides routines for manipulating matrices.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996).
! ------------------------------------------------------------------------
! Constants
! ------------------------------------------------------------------------
   USE TYPES                 ! predefined data-types
   IMPLICIT NONE
   INTERFACE swap
      MODULE PROCEDURE swap_i,swap_d,swap_dv,swap_dm, &
         masked_swap_ds,masked_swap_dv,masked_swap_dm
   END INTERFACE
   PRIVATE
   PUBLIC :: swap, ludcmp, lubksb, lubksbm, qrdcmp, qrsolv, qrdecomp, &
             v2abs, v1abs, m1abs, m1absinv, L1cond, outerprod
! ------------------------------------------------------------------------
! Subroutines
! ------------------------------------------------------------------------
CONTAINS

FUNCTION v2abs(v)
! Version 1.00, last modified 18-09-1998.
! --- Function and Algorithm ---
! L2 norm of vector.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1008.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: v2abs
   REAL(KIND = dp), INTENT(in) :: v(:)
!        v ... vector to find norm
! --- Code ---
   v2abs = SQRT(DOT_PRODUCT(v,v))
   RETURN
END FUNCTION v2abs

FUNCTION v1abs(v)
! Version 1.00, last modified 21-10-1998.
! --- Function and Algorithm ---
! L1 norm of vector.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: v1abs
   REAL(KIND = dp), INTENT(in) :: v(:)
!        v ... vector to find norm
! --- Code ---
   v1abs = SUM(ABS(v))
   RETURN
END FUNCTION v1abs

FUNCTION m1abs(m)
! Version 1.00, last modified 21-10-1998.
! --- Function and Algorithm ---
! L1 norm of matrix.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: m1abs
   REAL(KIND = dp), INTENT(in) :: m(:,:)
!        m ... matrix to find norm
! --- Code ---
   m1abs = MAXVAL(SUM(ABS(m),dim=2))
   RETURN
END FUNCTION m1abs

FUNCTION m1absinv(m)
! Version 1.00, last modified 21-04-1999.
! --- Function and Algorithm ---
! L1-norm of inverse of matrix.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: m1absinv
   REAL(KIND = dp), INTENT(in) :: m(:,:)
   REAL(KIND = dp), ALLOCATABLE :: Aqr(:,:), R(:,:)
   INTEGER(KIND = i4b), ALLOCATABLE :: luindx(:)
   INTEGER(KIND = i4b) :: n, i
   n = SIZE(m,1)

   ALLOCATE(Aqr(n,n),R(n,n),luindx(n))
   R = m
   Aqr = 0.0_dp              ! eye(dimp)
   DO i = 1, n
      Aqr(i,i) = 1.0_dp
   END DO
   CALL ludcmp(R,luindx)
   CALL lubksbm(R,luindx,Aqr)
   m1absinv = m1abs(Aqr)
   DEALLOCATE(luindx, R, Aqr)
   RETURN
END FUNCTION m1absinv

FUNCTION L1cond(m)
! Version 1.01, last modified 21-04-1999.
! --- Function and Algorithm ---
! L1-condition number of matrix.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp) :: L1cond
   REAL(KIND = dp), INTENT(in) :: m(:,:)
   L1cond = m1abs(m)*m1absinv(m)
   RETURN
END FUNCTION L1cond

!MODULAR SUBROUTINE swap(v1, v2)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Swap two arguments.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.991.
SUBROUTINE swap_i(a,b)
INTEGER(KIND = i4b), INTENT(in out) :: a,b
INTEGER(KIND = i4b) :: dum
   dum=a
   a=b
   b=dum
   RETURN
END SUBROUTINE swap_i
SUBROUTINE swap_d(a,b)
REAL(KIND = dp), INTENT(in out) :: a,b
REAL(KIND = dp) :: dum
   dum=a
   a=b
   b=dum
   RETURN
END SUBROUTINE swap_d
SUBROUTINE swap_dv(a,b)
REAL(KIND = dp), DIMENSION(:), INTENT(in out) :: a,b
REAL(KIND = dp), DIMENSION(SIZE(a)) :: dum
   dum=a
   a=b
   b=dum
   RETURN
END SUBROUTINE swap_dv
SUBROUTINE swap_dm(a,b)
COMPLEX(KIND = dp), DIMENSION(:,:), INTENT(in out) :: a,b
COMPLEX(KIND = dp), DIMENSION(size(a,1),size(a,2)) :: dum
   dum=a
   a=b
   b=dum
   RETURN
END SUBROUTINE swap_dm
SUBROUTINE masked_swap_ds(a,b,mask)
REAL(KIND = dp), INTENT(in out) :: a,b
LOGICAL, INTENT(IN) :: mask
REAL(KIND = dp) :: swp
   if (mask) then
      swp=a
      a=b
      b=swp
   end if
   RETURN
END SUBROUTINE masked_swap_ds
SUBROUTINE masked_swap_dv(a,b,mask)
REAL(KIND = dp), DIMENSION(:), INTENT(in out) :: a,b
LOGICAL, DIMENSION(:), INTENT(IN) :: mask
REAL(KIND = dp), DIMENSION(size(a)) :: swp
   where (mask)
      swp=a
      a=b
      b=swp
   end where
   RETURN
END SUBROUTINE masked_swap_dv
SUBROUTINE masked_swap_dm(a,b,mask)
REAL(KIND = dp), DIMENSION(:,:), INTENT(in out) :: a,b
LOGICAL, DIMENSION(:,:), INTENT(IN) :: mask
REAL(KIND = dp), DIMENSION(size(a,1),size(a,2)) :: swp
   where (mask)
      swp=a
      a=b
      b=swp
   end where
   RETURN
END SUBROUTINE masked_swap_dm

!SUBROUTINE swap(v1, v2)
!! Version 1.00, last modified 10-01-1998.
!! --- Function and Algorithm ---
!! Swap two vectors.
!! --- Bibliography ---
!! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
!! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.991.
!! --- Arguments ---
!   IMPLICIT NONE
!   REAL(KIND = dp), INTENT(in out) :: v1(:), v2(:)
!!        v1, v2 ... vectors to swap
!! --- Local variables ---
!   REAL(KIND = dp) :: dum(SIZE(v1))
!!        dum ... auxiliary data
!! --- Code ---
!   dum = v1
!   v1 = v2
!   v2 = dum
!   RETURN
!END SUBROUTINE swap

FUNCTION outerprod(v1, v2)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Calculate the matrix of the outer product of two vectors, optimised for
! parallel processing. Called by the lu and qr decomposition
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1000.
! --- Arguments ---
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: v1(:), v2(:)
!        v1, v2 ... vectors to calculate outer product
   REAL(KIND = dp) :: outerprod(SIZE(v1),SIZE(v2))
! --- Code ---
   outerprod = SPREAD(v1, dim=2, ncopies=SIZE(v2)) * &
               SPREAD(v2, dim=1, ncopies=SIZE(v1))
   RETURN
END FUNCTION outerprod

SUBROUTINE ludcmp(A, indx)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Replace matrix A by the LU-decomposition of a rowwise permutation.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1016.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in out) :: A(:,:)
!        A ... (in) matrix to decompose (out) LU-decomposition
   INTEGER(KIND = i4b), INTENT(out) :: indx(:)
!        indx ... row permutation by partial pivoting
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: vv(SIZE(A,1))
!        vv ... implicit scaling of each row
   INTEGER(KIND = i4b) :: j, n, imax
!        j ... loop counter
!        n ... dimension of system
!        imax ... pivot row
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   n = SIZE(A,1)
   IF ((n /= SIZE(A,2)) .OR. (n /= SIZE(indx)) ) THEN
!     error in input dimensions (not checked)
   END IF

   vv = MAXVAL(ABS(A), dim=2)
   IF ((ANY(vv == 0.0_dp)) .OR. &
       (ANY(vv < MAXVAL(vv)*EPSILON(1.0_dp))) ) THEN
      WRITE(*,*) ' singular matrix in LUDCMP '
      STOP
   END IF
   vv = 1.0_dp/vv
   DO j = 1, n
      imax = (j-1)+imaxloc(vv(j:n)*ABS(A(j:n,j))) ! find pivot row
      IF (j /= imax) THEN
         CALL swap(A(imax,:), A(j,:))
         vv(imax) = vv(j)
      END IF
      indx(j) = imax
      IF (A(j,j) == 0.0_dp) A(j,j) = TINY(1.0_dp)
      A(j+1:n,j) = A(j+1:n,j) / A(j,j) ! divide by pivot element
      A(j+1:n,j+1:n) = A(j+1:n,j+1:n) - outerprod(A(j+1:n,j), A(j,j+1:n))
   END DO
   RETURN
! ------------------------------------------------------------------------
! Local subroutines
! ------------------------------------------------------------------------
   CONTAINS

FUNCTION imaxloc(v)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Get location of largest element of vector.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.993.
! --- Arguments ---
   IMPLICIT NONE
   INTEGER(KIND = i4b) :: imaxloc
   REAL(KIND = dp), INTENT(in) :: v(:)
!        v ... vector to find location
! --- Local variables ---
   INTEGER(KIND = i4b) :: imax(1)
!        imax ... vector of length 1 to use MAXLOC
! --- Code ---
   imax = MAXLOC(v(:))
   imaxloc = imax(1)
   RETURN
END FUNCTION imaxloc
END SUBROUTINE ludcmp

SUBROUTINE lubksb(A, indx, b)
! Version 1.00, last modified 10-01-1998.
! --- Function and Algorithm ---
! Solve a linear equation system Ax=b with the LU-decomposition and
! permutation vector at hand (to be called after LUDCMP).
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1017.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: A(:,:)
!        A ... LU-decomposition matrix
   INTEGER(KIND = i4b), INTENT(in) :: indx(:)
!        indx ... row permutation by partial pivoting
   REAL(KIND = dp), INTENT(in out) :: b(:)
!        b ... (in) right side of LES, (out) solution of LES
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: summ
!        summ ... auxiliary data
   INTEGER(KIND = i4b) :: i, n, ii, ll
!        i ... loop counter
!        n ... dimension of system
!        ii, ll ... auxiliary indices
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   n = SIZE(A,1)
   IF ((n /= SIZE(A,2)) .OR. (n /= SIZE(indx)) .OR. (n /= SIZE(b)) ) THEN
!     error in input dimensions (not checked)
   END IF

   ii = 0
   DO i = 1, n
      ll = indx(i)
      summ = b(ll)
      b(ll) = b(i)
      IF (ii /= 0) THEN
         summ = summ - DOT_PRODUCT(A(i,ii:i-1), b(ii:i-1))
      ELSE IF (summ /= 0.0_dp) THEN
         ii = i
      END IF
      b(i) = summ
   END DO
   DO i = n,1,-1
      b(i) = (b(i) - DOT_PRODUCT(A(i,i+1:n), b(i+1:n)))/A(i,i)
   END DO
   RETURN
END SUBROUTINE lubksb

SUBROUTINE lubksbm(A, indx, B)
! Version 1.00, last modified 01-10-1998.
! --- Function and Algorithm ---
! Solve a matrix linear equation system AX=B with the LU-decomposition and
! permutation vector at hand (to be called after LUDCMP).
! --- Bibliography ---
! based on "lubksb" W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: A(:,:)
!        A ... LU-decomposition matrix
   INTEGER(KIND = i4b), INTENT(in) :: indx(:)
!        indx ... row permutation by partial pivoting
   REAL(KIND = dp), INTENT(in out) :: B(:,:)
!        b ... (in) right sides of the LES, (out) solution of the LES
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: summ
!        summ ... auxiliary data
   INTEGER(KIND = i4b) :: i, n, ii, ll, j
!        i ... loop counter
!        n ... dimension of system
!        ii, ll ... auxiliary indices
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   n = SIZE(A,1)
   IF ((n /= SIZE(A,2)) .OR. (n /= SIZE(indx)) .OR. (n /= SIZE(B,1))) THEN
!     error in input dimensions (not checked)
   END IF

   DO j = 1, SIZE(B,2)
      ii = 0
      DO i = 1, n
         ll = indx(i)
         summ = b(ll,j)
         b(ll,j) = b(i,j)
         IF (ii /= 0) THEN
            summ = summ - DOT_PRODUCT(A(i,ii:i-1), b(ii:i-1,j))
         ELSE IF (summ /= 0.0_dp) THEN
            ii = i
         END IF
         b(i,j) = summ
      END DO ! i = 1, n
      DO i = n,1,-1
         b(i,j) = (b(i,j) - DOT_PRODUCT(A(i,i+1:n), b(i+1:n,j)))/A(i,i)
      END DO ! i = 1, n
   END DO ! j = 1, SIZE(B,2)

   RETURN
END SUBROUTINE lubksbm

SUBROUTINE qrdcmp(A, c, d, sing)
! Version 1.00, last modified 18-09-1998.
! --- Function and Algorithm ---
! Calculate QR-decomposition of quadratic matrix A.
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1039.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in out) :: A(:,:)
!        A ... (in) nxn matrix to decompose (out) R from QR-decomposition
   REAL(KIND = dp), INTENT(out) :: c(:), d(:)
!        c ... factors of the n-1 Householder matrices building Q
!        d ... diagonal elements of R
   LOGICAL, INTENT(out) :: sing
!        sing ... true if singularity is encountered
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: scale, sigma
!        scale ... scale of lower part of a collumn
!        sigma ... sign
   INTEGER(KIND = i4b) :: j, n
!        j ... loop counter
!        n ... dimension of system
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   n = SIZE(A,1)
   IF ( (n /= SIZE(A,2)) .OR. (n /= SIZE(c)) .OR. (n /= SIZE(d)) ) THEN
!     error in input dimensions (not checked)
   END IF
   sing = .false.
   c = 0.0_dp
   d = 0.0_dp

   DO j = 1,n-1
      scale = MAXVAL(ABS(A(j:n,j)))
      IF ( (scale == 0.0_dp) .OR. &
           (scale < MAXVAL(ABS(A))*EPSILON(1.0_dp)) ) THEN ! singular case
         sing = .true.
         c(j) = 0.0_dp
         d(j) = 0.0_dp
      ELSE                   ! form Qj and Qj A
         A(j:n,j) = A(j:n,j) / scale
         sigma = SIGN(v2abs(A(j:n,j)),A(j,j))
         A(j,j) = A(j,j) + sigma
         c(j) = sigma * A(j,j)
         d(j) = -scale * sigma
         A(j:n,j+1:n) = A(j:n,j+1:n) - outerprod(A(j:n,j), &
                        MATMUL(A(j:n,j), A(j:n,j+1:n))) / c(j)
      END IF
   END DO
   d(n) = A(n,n)
   IF ( (d(n) == 0.0_dp) .OR. &
        (ABS(d(n)) < MAXVAL(ABS(A))*EPSILON(1.0_dp)) ) sing = .true.
   RETURN
END SUBROUTINE qrdcmp

SUBROUTINE qrsolv(A, c, d, b)
! Version 1.00, last modified 18-09-1998.
! --- Function and Algorithm ---
! Solve a linear *regular* equation system Ax=b with QR-decomposition
! (to be called after QRDCMP).
! --- Bibliography ---
! W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery,
! 'Numerical Recipes in Fortran 90', Cambridge Univ. Press (1996), p.1040.
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: A(:,:), c(:), d(:)
!        A ... R from QR-decomposition
!        c ... factors of the n-1 Householder matrices building Q
!        d ... diagonal elements of R
   REAL(KIND = dp), INTENT(in out) :: b(:)
!        b ... (in) right side of LES, (out) solution of LES
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: tau
!        tau ... auxiliary factor
   INTEGER(KIND = i4b) :: j, n
!        j ... loop counter
!        n ... dimension of system
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   n = SIZE(A,1)
   IF ((n /= SIZE(A,2)) .OR. (n /= SIZE(c)) .OR. (n /= SIZE(d)) &
                        .OR. (n /= SIZE(b))) THEN
!     error in input dimensions (not checked)
   END IF

   DO j = 1,n-1              ! form, Q^T b
      tau = DOT_PRODUCT(A(j:n,j), b(j:n)) / c(j)
      b(j:n) = b(j:n) - tau * A(j:n,j)
   END DO
   b(n) = b(n) / d(n)
   DO j = n-1,1,-1
      b(j) = (b(j) - DOT_PRODUCT(A(j,j+1:n), b(j+1:n))) / d(j)
   END DO
   RETURN
END SUBROUTINE qrsolv

SUBROUTINE qrdecomp(A, Q)
! Version 1.01, last modified 22-09-1998.
! --- Function and Algorithm ---
! Calculate the QR-decomposition of nxm matrix A.
! --- Bibliography ---
! based on 'qrdcmp' W.H.Press, S.A.Teukolsky, W.T.Vetterling, B.P.Flannery
! ------------------------------------------------------------------------
! Arguments
! ------------------------------------------------------------------------
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in out) :: A(:,:)
!        A ... (in) nxm matrix to decompose (out) nxm matrix R
   REAL(KIND = dp), INTENT(out) :: Q(SIZE(A,1),SIZE(A,1))
!        Q ... nxn matrix Q from QR-decomposition
! ------------------------------------------------------------------------
! Local variables
! ------------------------------------------------------------------------
   REAL(KIND = dp) :: Qj(SIZE(A,1),SIZE(A,1)), u(SIZE(A,1))
!        Qj ... householder matrices
!        u ... vector axis of householder matrix
   REAL(KIND = dp) :: scale, sigma, c
!        scale ... scale of lower part of a collumn
!        sigma ... sign
!        c ... factor of the Householder matrix building Q
   INTEGER(KIND = i4b) :: i, j, n, m, k
!        i, j ... loop counter
!        n, m, k ... dimension of system
! ------------------------------------------------------------------------
! Code
! ------------------------------------------------------------------------
   n = SIZE(A,1)
   m = SIZE(A,2)
   k = MIN(n-1,m)
   Q = 0.0_dp                ! set up Q as product of Householder matrices
   DO i = 1,n
      Q(i,i) = 1.0_dp
   END DO

   DO j = 1,k
      scale = MAXVAL(ABS(A(j:n,j)))
!      IF ( (scale == 0.0_dp) .OR. &
!           (scale < MAXVAL(ABS(A))*EPSILON(1.0_dp)) ) THEN ! singular case
      IF (scale == 0.0_dp) THEN ! singular case
         A(j:n,j) = 0.0_dp   ! calculate R
      ELSE
         IF (j > 1) u(1:j-1) = 0.0_dp ! axis for Householder matrix
         u(j:n) = A(j:n,j) / scale
         sigma = SIGN(v2abs(u(j:n)),u(j))
         u(j) = u(j) + sigma
         c = sigma * u(j)

         Qj = 0.0_dp         ! generate Q_j
         DO i = 1,n
            Qj(i,i) = 1.0_dp
         END DO
         Qj(j:n,j:n) = Qj(j:n,j:n) - outerprod(u(j:n),u(j:n)) / c
         Q = MATMUL(Q,Qj)    ! accumulate Q_j into Q

         A(j:n,j+1:m) = A(j:n,j+1:m) - outerprod(u(j:n), &
                        MATMUL(u(j:n), A(j:n,j+1:m))) / c ! update A
         A(j+1:n,j) = 0.0_dp ! calculate R
         A(j,j) = -scale * sigma
      END IF
   END DO
!   IF (k == n-1) A(n,n) = A(n,n)
   RETURN
END SUBROUTINE qrdecomp

END MODULE linalg
