PROGRAM SSHT
! Single SHooTing with implicit Euler IDeC method
! ----------------------------------------------------------------------
! Version 1.01, last modified 18-06-1999.
! --- Function ---
! Do single shooting for singular BVPs with separated BCs by using a
! black-box (module "ieuleridcs", subroutine "ivp_ieuleridec") and
! reduced superposition. Non-linear problems are solved with a damped
! Newton method. The code has to know the exact basis of the kernel
! of M, which is generated automatically. First we initialise all
! variables, then we branch for linear or non-linear case and last
! we calculate the whole solution.
! --- Author ---
! Peter W. KOFLER,
! Department of Applied Mathematics and Numerical Analysis,
! Vienna University of Technology, Vienna, Austria.
! ----------------------------------------------------------------------
   USE types                 ! data types double and long integer
   USE linalg                ! linear algebra for LU decomposition
   USE ieulidcs              ! implicit Euler with IDeC IVP solver as
                             ! Black Box for the shooting algorithm
   USE funbox                ! user input: the BVP of ODE
   USE bigfun                ! big function for simulates integration
! ----------------------------------------------------------------------
!        Constants
! ----------------------------------------------------------------------
   IMPLICIT NONE
   INTEGER(KIND = i4b), PARAMETER :: neqn = box_neqn
!        neqn ... dimension of system of ODEs
   INTEGER(KIND = i4b), PARAMETER :: max_newton_steps = 50, &
                                     max_lambda_steps = 20
!        max_newton_steps ... maximum of steps for Newton-iteration used
!        max_lambda_steps ... max. of damping steps in Newton-method
   REAL(KIND = dp), PARAMETER :: dumping_factor_newton = 1.0_dp, &
                                 maxinitstart = 1.0D-10, &
                                 maxinitmul = 1.2_dp ! 40 with 2.0,
!        dumping_factor_ ... factor of first dumped Newton iteration
!        maxinitstart ... start value of error in estimation of IV s
!        maxinitmul ... multiplicate maxinit by this for next loop
   CHARACTER(LEN = 12) :: name1 = 'log0.nwt',& ! [iter, |res|, |incr|]
                          name2 = 'log0.cnv',& ! [maxinit,maxit,glob.F.]
                          name4 = 'log0.ord'   ! [|res|, p, c]
   LOGICAL :: debug = .true., & ! we know the solution
              verbose = .false. ! do some output
   INTEGER(KIND = i4b), PARAMETER :: initloop = 0 ! 150
!        initloop ... number of loops duplicating maxinitial
! ----------------------------------------------------------------------
!        Local variables
! ----------------------------------------------------------------------
! ========= properties of the BVP =========
!        neqn (integer,fun) ... dimension of system (USER)
!        linear (logical,fun) ... problem is linear (USER)
   REAL(KIND = dp) :: a, b
!        a,b ... interval to perform shooting (USER)
   REAL(KIND = dp) :: Ba(neqn,neqn), Bb(neqn,neqn), beta(neqn)
!        Ba*y(a)+Bb*y(b)=beta ... linear BCs of BVP (USER)
   REAL(KIND = dp) :: y0(neqn), y1(neqn)
!        y0, y1 ... y(a) and y(b) for testing (debug)
   INTEGER(KIND = i4b) :: dimq, dimp
!        dimq ... number of separated ICs at a = dimension of span(Q)
!        dimp ... complement number n-p = dimension of kernel(M)
! ========= properties of the black-box IVP solver =========
   REAL(KIND = dp) :: basictol, tol, eps, h, h2, hmin
!        basictol ... global error tolerance for IVP solver
!        tol ... global error tolerance and global error estimation
!        eps ... Newton residual tolerance (only non-linear problems)
!        h, hmin ... step size and smallest allowed step size
!        h2 ... step size calculating Jacobian
   INTEGER(KIND = i4b) :: ifail, maxidec, pord
!        ifail ... error flag for black-box
!        maxixdec ... number of IDeC iterations
!        pord ... order of Zadunaisky polynomial
   REAL(KIND = dp), ALLOCATABLE :: Z(:)
!        Z ... IVs for big system of ICs
   REAL(KIND = dp) :: y(neqn), maxerr, maxerrx
!        y ... IVs for last integration (linear)
!              actual initial value (non-linear)
!        maxerr ... maximum of global error on [a,b] (debug)
!        maxerrx ... position of this maxerr (debug)
! ========= variables for linear case (reduced superposition) =========
   REAL(KIND = dp), ALLOCATABLE :: Aqr(:,:), R(:,:), Q(:,:)
!        Aqr ... matrix to QR decompose the separated part of Ba
!        Q,R ... QR decomposition of (p times n) IVs
   REAL(KIND = dp), ALLOCATABLE :: grY0(:,:), grY1(:,:), sbeta(:)
!        grY0 ... ICs homogenous nxq problem = fundamental matrix in the
!                 linear case = that is the kernel of M
!        grY1 ... result of homogenous nxq problem at b
!        sbeta ... right hand side of LES Q*s=sbeta
   REAL(KIND = dp) :: ya(neqn), yb(neqn)
!        ya ... ICs inhomogenous problem (special solution) (linear)
!               iterative solution at a gained by shooting (non-linear)
!        yb ... result inhomogenous problem
!               iterative solution at b gained by shooting (non-linear)
   INTEGER(KIND = i4b), ALLOCATABLE :: luindx(:)
!        luindx ... index for pivoting LU-decomposition
! ========= variables for non-linear case (Newton Iteration) =========
   REAL(KIND = dp) :: nrmF, nrmds, nrmo1ds, nrmo2ds, dsp, dsc, Beps, &
                      lambda, sqrteps, maxinitial
!        nrmF ... norm of residual |F(s)|
!        nrmds ... norm of Newton increment |ds|
!        nrmo1ds, nrmo2ds ... norm of old Newton increment |ds-1|,|ds-2|
!        dsp, dsc ... order and order constant of Newton increments
!        Beps ... Newton epsilon = tolerance to residual
!        lambda ... last damping factor
!        sqrteps ... delta for numerically differentiating (non-linear)
!        maxinitial ... relative difference to exact initial value
   REAL(KIND = dp), ALLOCATABLE :: s(:), F(:), deltaF(:,:), deltas(:), &
                                   delta(:), Za(:), Ze(:), slambda(:)
!        s ... linear combination of allowed IVs
!        F ... residual, right hand side of LES F'(s)*ds=-F(s)
!        deltaF ... F'(s), Jakobian of Newton's method
!        deltas ... increment
!        delta ... vector of differential quotients for Jacobian
!        Za ... IVs of big system calculating Jacobian
!        Ze ... EVs of big system calculating Jacobian
!        slambda ... test value for damping
   INTEGER :: iter, lindex
!        iter ... number of Newton iteration
!        lindex ... iteration count in damping loop
! ========= misc. =========
   INTEGER(KIND = i4b) :: i, j, k
!        i,j,k ... loop counter
   
! ----------------------------------------------------------------------
!        Code
! ----------------------------------------------------------------------
   WRITE(*,*) 'Single Shooting (implicit Euler Method with IDeC)'

   ! initialise the BVP
   CALL box_bcond(a, b, Ba, Bb, beta)
   y0 = box_true(a,y0)       ! get y(a) and y(b) for testing accuracy
   y1 = box_true(b,y1)

   IF (box_linear) THEN
     WRITE(*,'(I3,A,F6.4,A,F6.4)') neqn,'-dimensional linear BVP on ', &
                                   a,' to ',b
   ELSE
     WRITE(*,'(I3,A,F6.4,A,F6.4)') neqn, &
                           '-dimensional non-linear BVP on ', a,' to ',b
   END IF

   ! we have Ba*y(a)+Bb*y(b)=beta, unsorted, now look for separated IVs
   dimq = 0
   DO i = 1, neqn
      IF (ALL(Bb(i,:) == 0.0_dp)) dimq = dimq + 1
   END DO ! i = 1, neqn
   dimp = neqn-dimq          ! complement dimension
   WRITE(*,'(A,I2,A,I2,A,I2,A)') ' dimension',neqn,' =',dimq, &
                                ' (separated) +',dimp,' (not separated)'

   ! swap lines to get these separated lines as the first dimq ones in
   ! the upper part of Ba and Bb.
   IF (dimq>0) THEN          ! there are separated BCs, so sort them
      i = 1
      j = neqn
      DO          
        IF ((.not.ALL(Bb(i,:)==0.0_dp)).AND.(ALL(Bb(j,:)==0.0_dp))) THEN
           CALL swap(Ba(i,:),Ba(j,:)) ! swap lines of linear BC matrices
           CALL swap(Bb(i,:),Bb(j,:))
           CALL swap(beta(i),beta(j))
           i = i + 1         ! the new lines are ok
           j = j - 1
        ELSE IF (ALL(Bb(i,:) == 0.0_dp)) THEN
           i = i + 1         ! these lines are ok
        ELSE IF (.not. (ALL(Bb(i,:) == 0.0_dp))) THEN
           j = j -1
        END IF
        IF (i .ge. j) EXIT   ! all lines checked
      END DO ! sort BCs
   END IF ! (dimq>0)

   IF (verbose) THEN
      WRITE(*,*) '**** the sorted BC (Ba*y(a)+Bb*y(b)=beta) are ****'
      WRITE(*,*) 'Ba'
      CALL write_mat(Ba)
      WRITE(*,*) 'Bb'
      CALL write_mat(Bb)
      WRITE(*,*) 'beta'
      WRITE(*,*) beta
   END IF

! ----------------------------------------------------------------------
!        initialise the IVP Solver
! ----------------------------------------------------------------------
   log_file = .false.
   bdebug = debug
   sqrteps = SQRT(EPSILON(1.0_dp))

!        number of IDeC iterations in black-box
!   READ(*,*) maxidec        ! FPR dummy read
   maxidec = 4
   WRITE(*,'(A,I1)') ' number of IDeC iterations in black-box: ',maxidec

!        order of method and polynomial degree in black-box
   WRITE(*,'(A,I1)') ' order of black-box method: ', maxidec+1
   pord = maxidec+1
   WRITE(*,'(A,I1)') ' degree of Zadunaisky-polynomials: ', pord

!        global error tolerance
   IF (debug) THEN
      WRITE(*,*) 'global error control in black-box deactivated'
      basictol = 1.0D+99     ! set high tolerance to accept any solution
                             ! to investigate order of convergence
   ELSE
      WRITE(*,*) 'tolerance to global error in black-box:'
      basictol = 1.0D-9
!     READ(*,*) basictol
      WRITE(*,*) basictol
   END IF

!        starting step size
   WRITE(*,*) 'starting step size(H):'
   hmin = 0.0_dp
   READ(*,*) h
   h = ABS(h)*(b-a)
   h2 = h
   WRITE(*,*) ABS(h)/pord    ! real step size = h/pord

   IF (.not. box_linear) THEN ! non-linear BVP needs Newton iterations
      WRITE(*,*) 'Newton tolerance in black-box:'
      eps = basictol / 10.0_dp
      eps = 1.0D-14
!     READ(*,*) eps
      WRITE(*,*) eps

      WRITE(*,*) 'Newton tolerance for shooting procedure:'
      beps = basictol
      beps = eps * 100.0_dp
!     READ(*,*) Beps
      WRITE(*,*) Beps

!        Estimation of IV-Eps
      WRITE(*,*) 'relative error of estimated initial value:'
      maxinitial = 0.0_dp
      IF (initloop == 0) maxinitial = maxinitstart*5.0_dp**13
!     READ(*,*) maxinitial
      WRITE(*,*) maxinitial
   ELSE
      eps = 0.0_dp
      beps = 0.0_dp
      maxinitial = 0.0_dp
   END IF ! (.not. box_linear)

   IF (initloop > 0) OPEN(15, FILE = name2)
   DO k = 0, initloop

! ----------------------------------------------------------------------
!        calculate the IVs for reduced superposition by QR decomposition
! ----------------------------------------------------------------------
   WRITE(*,*) 'calculating initial value'
   IF (dimq > 0) THEN     ! reduced superposition is possible
      ALLOCATE(grY0(neqn,dimp),R(dimq,dimq),Q(neqn,neqn),Aqr(neqn,dimq))
      ! calculate complement to separated ICs, we just need to shoot for
      ! this vectors: Phi bar = back part of Q^(-1)^T = Q
      Aqr = transpose(Ba(1:dimq,:))
      CALL qrdecomp(Aqr, Q)  ! yields Q in Q and R in Aqr
      grY0 = Q(:,dimq+1:neqn)
      R = Aqr(1:dimq,:)      ! we just need the upper part of R
      DEALLOCATE(Aqr)

      ! calculate inverse of R as we need R^(-1)^T for the special IV
      IF (dimq > 1) THEN
         WRITE(*,*) 'L1-condition of R in Q.R of Ba:',L1cond(R)
         ALLOCATE(Aqr(dimq,dimq),luindx(dimq))
         Aqr = 0.0_dp        ! eye(dimq)
         DO i=1,dimq
            Aqr(i,i) = 1.0_dp
         END DO
         CALL ludcmp(R,luindx)
         CALL lubksbm(R,luindx,Aqr)
         ya = MATMUL(MATMUL(Q(:,1:dimq),transpose(Aqr)),beta(1:dimq))
         DEALLOCATE(luindx, Aqr)
      ELSE
         ya = Q(:,1)*beta(1)/R(1,1) ! 1-dimensional case
      END IF ! (dimq > 1)
      DEALLOCATE(Q, R)
   ELSE ! (dimq>0)
      ALLOCATE(grY0(neqn,neqn)) ! no separated lines
      grY0 = 0.0_dp
      DO i=1,dimp            ! eye(dimq)
         grY0(i,i) = 1.0_dp
      END DO
      ya = 0.0_dp            ! we have to start with 0 as special IV
   END IF ! (dimq > 0)

   IF (verbose) THEN
      WRITE(*,*) '**** the initial values are ****'
      WRITE(*,*) 'grY0 (fundamental matrix/kernel)'
      CALL write_mat(grY0)
      WRITE(*,*) 'ya (special value)'
      WRITE(*,*) ya
   END IF

   CALL big_init(dimp)       ! make dimensions available to other module
   IF (box_linear) THEN      ! linear case
! ----------------------------------------------------------------------
!        single shooting for linear problems
! ----------------------------------------------------------------------

      ! solve for Y(b)=grY1 (nxq) and y(b)=yb (n)
      ALLOCATE(grY1(neqn,dimp),Z((dimp+1)*neqn))
      DO i = 1, dimp
         Z((i-1)*neqn+1:i*neqn) = grY0(:,i)
      END DO ! DO i over diagonal
      Z(dimp*neqn+1:(dimp+1)*neqn) = ya ! Z contains IVs as long system

      tol = basictol         ! adjust global error tolerance
      ifail = 0
      CALL ivp_ieuleridec(bigl_fcn, a, Z, b, h, hmin, tol, eps, &
           box_linear, maxidec, pord, ifail, bigl_dfcn, dummy1, dummy2)
      IF (ifail /= 0) CALL error(ifail,.true.)
      IF (verbose) WRITE(*,*) 'used stepsize: ', h/pord

      ! transform the results
      grY1 = 0.0_dp
      yb = 0.0_dp
      DO i = 1, dimp
         grY1(:,i) = Z((i-1)*neqn+1:i*neqn)
      END DO ! DO i over diagonal
      yb = Z(dimp*neqn+1:(dimp+1)*neqn)
      DEALLOCATE(Z)

      ! set up LES Q*s=sbeta and calculate missing IVs "s" (in sbeta)
      ALLOCATE(sbeta(dimp), Q(dimp,dimp), luindx(dimp))
      Q = MATMUL(Ba(dimq+1:neqn,:),grY0)+MATMUL(Bb(dimq+1:neqn,:),grY1)
      sbeta = beta(dimq+1:neqn) - MATMUL(Ba(dimq+1:neqn,:),ya) - &
                                  MATMUL(Bb(dimq+1:neqn,:),yb)
      IF (dimp > 1) THEN
         WRITE(*,*) 'L1-condition Q:',L1cond(Q)
         CALL ludcmp(Q,luindx)
         CALL lubksb(Q,luindx,sbeta)
      ELSE
         sbeta = sbeta/Q(1,1) ! 1-dimensional case
      END IF ! (dimp > 1)
      DEALLOCATE(luindx, Q)

      ! calculate solution with starting value "s" (in sbeta)
      ya = MATMUL(grY0,sbeta)+ya
      yb = MATMUL(grY1,sbeta)+yb
      DEALLOCATE(sbeta,grY1,grY0)

      IF (verbose) THEN
         WRITE(*,*) 'solution at a is         (should be)'
         DO i = 1, neqn
            WRITE(*,*) ya(i), '(',y0(i),')'
         END DO
         WRITE(*,*) 'with global error ', MAXVAL(ABS(ya-y0))

         WRITE(*,*) 'solution at b is         (should be)'
         DO i = 1, neqn
            WRITE(*,*) yb(i), '(',y1(i),')'
         END DO
         WRITE(*,*) 'with global error ', MAXVAL(ABS(yb-y1))
      END IF
   ELSE                      ! non-linear case
! ----------------------------------------------------------------------
!        single shooting for non-linear problems by Newton's method
! ----------------------------------------------------------------------

      ! estimate an initial value
      ALLOCATE(s(dimp), F(dimp), deltas(dimp), deltaF(dimp,dimp), &
               luindx(dimp), Za(dimp*neqn), Ze(dimp*neqn), &
               delta(dimp), slambda(dimp) )
      s = 0.0_dp             ! estimate unknown components of y(a)
                             ! this estimation may be very bad.
      if (debug) THEN
         ! calculate exact s to manage amount of wrong estimation
         IF (dimq > 0) THEN  ! reduced superposition is possible
            ALLOCATE(Q(neqn,neqn),Aqr(neqn,dimq))
            Aqr = transpose(Ba(1:dimq,:))
            CALL qrdecomp(Aqr, Q) ! yields Q in Q and R in Aqr
            y = MATMUL(transpose(Q),y0-ya)
            s = y(dimq+1:neqn)
            DEALLOCATE(Aqr,Q)
         ELSE
            s = y0-ya
         END IF
         DO i = 1,dimp
            IF (s(i) /= 0.0_dp ) THEN
               s(i) = s(i)*(1.0_dp+maxinitial)
            ELSE
               s(i) = maxinitial
            END IF
         END DO ! i = 1,dimp
      END IF

      y = ya                 ! get fixed IVs forced by Ba*y(a)
      ya = y + MATMUL(grY0,s) ! fixed value and variable value not
                             ! forced by Ba*y(a), the IV

      IF (MAXVAL(ABS(y0)) <= EPSILON(1.0_dp) ) THEN
         WRITE(*,*) 'absolute error in initial value:', &
                    MAXVAL(ABS(y0-ya))
      ELSE
         WRITE(*,*) 'relative error in initial value:', &
                    MAXVAL(ABS(y0-ya))/MAXVAL(ABS(y0))
      END IF

      ! do damped Newton's method to find solution of non linear eqs
      lambda = dumping_factor_newton/2.0_dp ! initial damping factor

      yb = ya                ! calculate the 1st Newton-residual F(s)
      tol = basictol         ! adjust global error tolerance
      ifail = 0
      CALL ivp_ieuleridec(box_fcn, a, yb, b, h, hmin, tol, eps, &
           box_linear, maxidec, pord, ifail, box_dfcn, dummy1, dummy2)
      IF (ifail /= 0) CALL error(ifail,.true.)
      IF (verbose) WRITE(*,*) 'used stepsize: ', h/pord
      F = -MATMUL(Ba(dimq+1:neqn,:),ya) - MATMUL(Bb(dimq+1:neqn,:),yb) &
          + beta(dimq+1:neqn)
      IF (verbose) WRITE(*,*) '1st Newton Residual F(s)', F
      nrmF = MAXVAL(ABS(F))  ! norm of residual

!     IF (initloop>0) name1(4:4) = char(k+64)
      IF (initloop == 0) OPEN(14, FILE = name1) ! [iter, |res|, |incr|]
!     IF (initloop>0) name4(4:4) = char(k+64)
      IF (initloop == 0) OPEN(17, FILE = name4) ! [|res|, p, c]

      iter = 0
      ! start Newton's Iteration
      DO
         iter = iter + 1
         WRITE(*,*) iter,'. Newton iteration'

         ! solve for F'(s) (qxq) using discrete differential quotients
         DO i = 1, dimp
            delta(i) = MAX(1.0_dp,ABS(s(i)))*sqrteps
            Za((i-1)*neqn+1:i*neqn) = ya + delta(i)*grY0(:,i)
         END DO
         Ze = Za
         tol = basictol      ! adjust global error tolerance
         ifail = 0
         CALL ivp_ieuleridec(bign_fcn, a, Ze, b, h2, hmin, tol, eps, &
              box_linear, maxidec, pord, ifail, bign_dfcn, dummy1, &
              dummy2)
         IF (ifail /= 0) CALL error(ifail,.true.)
         IF (verbose) WRITE(*,*) 'used stepsize: ', h2/pord
         DO i = 1, dimp
            deltaF(:,i) = - MATMUL(Ba(dimq+1:neqn,:), &
                                              Za((i-1)*neqn+1:i*neqn)) &
                          - MATMUL(Bb(dimq+1:neqn,:), &
                                              Ze((i-1)*neqn+1:i*neqn)) &
                          + beta(dimq+1:neqn)
!           WRITE(*,*) 'Residual F(s+deltai)', deltaF(:,i)
            deltaF(:,i) = (deltaF(:,i) - F)/delta(i)
!           WRITE(*,*) 'DeltaF (column)', deltaF(:,i)
         END DO ! quotients

         ! set up LES deltaF*ds=-F and calculate Newton increment deltas
         deltas = F
         IF (dimp > 1) THEN
            WRITE(*,*) "L1-condition F'(s):",L1cond(deltaF)
            CALL ludcmp(deltaF,luindx)
            CALL lubksb(deltaF,luindx,deltas)
         ELSE
            deltas(1) = deltas(1)/deltaF(1,1) ! 1-dimensional case
         END IF
         nrmds = MAXVAL(ABS(deltas)) ! norm of Newton increment

         ! if we have enough values, calculate order of Newton's method
         IF ((initloop == 0).AND.(iter > 2).AND.(nrmds/=0.0_dp).AND.&
             (nrmo1ds/=0.0_dp)) THEN
            IF ((nrmo2ds/nrmo1ds-1.0_dp) /= 0.0_dp) THEN
               dsp = log(nrmo1ds/nrmds)/log(nrmo2ds/nrmo1ds)
               dsc = nrmds/(exp(log(nrmo1ds)*dsp))
            ELSE
               dsp = 0.0_dp
               dsc = 0.0_dp
            END IF
         ELSE
            dsp = 0.0_dp
            dsc = 0.0_dp
         END IF
         nrmo2ds = nrmo1ds
         nrmo1ds = nrmds     ! store old increment length
!        WRITE(*,*) 'Newton increment deltas', deltas

         ! exit iteration if accurate
         WRITE(*,*) 'Newton residual: ',nrmF
         WRITE(*,*) 'Newton increment:',nrmds
         IF (initloop == 0) WRITE(14, *) iter, nrmF, nrmds
         IF (initloop == 0) WRITE(17, *) nrmF, dsp, dsc ! [|res|, p, c]
         
         IF ((nrmF<Beps).and.(nrmds<Beps*(1.0_dp+nrmF))) EXIT

         ! update solution and iterate damping factors
         lambda = MIN(1.0_dp,lambda*2.0_dp)
         lindex = 0
         DO
            IF (lindex>0) WRITE(*,*) lindex,'. damped Newton iteration'
            lindex = lindex + 1
            slambda = s - deltas*lambda
            ya = y + MATMUL(grY0,slambda)

            ! calculate new Newton-residual
            yb = ya          ! calculate the 1st Newton-residual F(s)
            tol = basictol   ! adjust global error tolerance
            ifail = 0
            CALL ivp_ieuleridec(box_fcn, a, yb, b, h, hmin, tol, eps, &
                 box_linear, maxidec, pord, ifail, box_dfcn, dummy1, &
                 dummy2)
            IF (ifail /= 0) CALL error(ifail,.false.)
            IF (verbose) WRITE(*,*) 'used stepsize: ', h/pord
            IF (ifail == 0) THEN
               F = -MATMUL(Ba(dimq+1:neqn,:),ya) - &
                    MATMUL(Bb(dimq+1:neqn,:),yb) + beta(dimq+1:neqn)
               ! check "monotony test" of the residuals by increments
               deltas = F    ! simplified Newton-increment
               IF (dimp > 1) THEN
                  CALL lubksb(deltaF,luindx,deltas)
               ELSE
                  deltas(1) = deltas(1)/deltaF(1,1) ! 1-dimensional case
               END IF
               IF (verbose) WRITE(*,*) MAXVAL(ABS(deltas)), ' <= ', &
                                     (1.0_dp-lambda/2.0_dp)*nrmds
               IF (MAXVAL(ABS(deltas)) <=  &
                                      (1.0_dp-lambda/2.0_dp)*nrmds) EXIT
               ! "monotony test" is true therefore take this solution
            END IF ! (ifail == 0)

            ! shorten increment's length by 2
            lambda = lambda/2.0_dp
            IF (lindex > max_lambda_steps) THEN
               ifail = 6     ! error, residual did not decrease
               CALL error(ifail,.true.)
            END IF
         END DO ! lambda dumping
         s = slambda
         nrmF = MAXVAL(ABS(F)) ! norm of residual

        IF (verbose) THEN
            WRITE(*,*) 'solution at a is         (should be)'
            DO i = 1, neqn
               WRITE(*,*) ya(i), '(',y0(i),')'
            END DO
            WRITE(*,*) 'with global error ', MAXVAL(ABS(ya-y0))

            WRITE(*,*) 'solution at b is         (should be)'
            DO i = 1, neqn
               WRITE(*,*) yb(i), '(',y1(i),')'
            END DO
            WRITE(*,*) 'with global error ', MAXVAL(ABS(yb-y1))
        END IF

         IF (iter > max_newton_steps) THEN
            ifail = 5        ! error, too many iterations
            CALL error(ifail,.true.)
         END IF
      END DO ! Do Newton iterations
      IF (initloop == 0) CLOSE(14)
      IF (initloop == 0) CLOSE(17)

      DEALLOCATE(slambda, delta, Ze, Za, luindx, deltaF, &
                 deltas, F, s, grY0)
      WRITE(*,*) 'number of Newton iterations:', iter
   END IF ! (box_linear)

! ----------------------------------------------------------------------
!        repeat integration to calculate complete solution curve
! ----------------------------------------------------------------------
   WRITE(*,*) 'calculating complete solution on [a,b]'
   tol = basictol            ! adjust global error tolerance
   ifail = 0
   y = ya                    ! set IVs

!  IF (initloop == 0) log_file = .true. ! enable plots
   CALL ivp_ieuleridec(box_fcn, a, y, b, h, hmin, tol, eps, box_linear,&
        maxidec, pord, ifail, box_dfcn, box_true, box_truelocal, &
        maxerr=maxerr, maxerrx=maxerrx)
   IF (ifail /= 0) CALL error(ifail,.true.)
   WRITE(*,*) 'used stepsize: ', h/pord

   IF (initloop > 0) THEN
      WRITE(*,*) 'the maximum of the global error is ', maxerr
      WRITE(15, *) maxinitial, iter, maxerr
      IF (k == 0 ) THEN
         maxinitial = maxinitstart
      ELSE
         maxinitial = maxinitial * maxinitmul
      END IF
      IF (k < initloop) THEN
         WRITE(*,*) 'relative error of estimated initial values:'
         WRITE(*,*) maxinitial
      END IF
   END IF
   END DO ! k=0, initloop
   IF (initloop > 0) CLOSE(15)

   WRITE(*,*) 'solution at a is         (should be)'
   DO i = 1, neqn
      WRITE(*,*) ya(i), '(',y0(i),')'
   END DO
   WRITE(*,*) 'with global error ', MAXVAL(ABS(ya-y0))

   WRITE(*,*) 'solution at b is         (was,                     ', &
              ' should be)'
   DO i = 1, neqn
      WRITE(*,*) y(i), '(', yb(i), ',', y1(i),')'
   END DO
   IF (.not. debug) WRITE(*,*) 'with global error estimation', tol
   WRITE(*,*) 'with global error ', MAXVAL(ABS(y-y1))
!  WRITE(*,*) 'and difference to first approximation ',MAXVAL(ABS(y-yb))

   WRITE(*,*) 'the maximum of the global error is ', maxerr
   WRITE(*,*) 'and it is at the position ' ! ,maxerrx
   WRITE(*,*) maxerrx

   WRITE(*,*) 'FUER ORDALL'
!  order of error |y(a)| and |y(b)|:
!  WRITE(*,*) MAXVAL(ABS(ya-y0)), MAXVAL(ABS(yb-y1))
!  order of error |y(b)| and |y([a,b])|:
!  WRITE(*,*) MAXVAL(ABS(y-y1)), maxerr
!  order of error |y(a)| and |y([a,b])|:
   WRITE(*,*) MAXVAL(ABS(ya-y0)), maxerr

   STOP

   CONTAINS
! ----------------------------------------------------------------------
SUBROUTINE write_mat(A)
! print a matrix for debugging purposes
   IMPLICIT NONE
   REAL(KIND = dp), INTENT(in) :: A(:,:)
   INTEGER(KIND = i4b) :: i
   DO i = 1, SIZE(A,1)
      IF (SIZE(A,1)>6) THEN
         WRITE(*,*) MAXVAL(ABS(A(i,:)))
      ELSE
         WRITE(*,*) A(i,:)
      END IF
   END DO
   RETURN
END SUBROUTINE write_mat
! ----------------------------------------------------------------------
SUBROUTINE error(ifail, bstop)
! write out error message and stop program
   IMPLICIT NONE
   INTEGER(KIND = i4b), INTENT(in) :: ifail
!        ifail ... error variable, set due to several errors
   LOGICAL, INTENT(in) :: bstop
!        bstop ... stop program
   IF ((ifail > -9).AND.(ifail < 5)) THEN
      WRITE(*,'(A,I1,A)') ' ERROR ',ifail,' in black-box:'
   ELSE
      WRITE(*,'(A,I1,A)') ' ERROR ',ifail,' in shooting procedure:'
   END IF
   SELECT CASE (ifail)
   CASE (0)
      WRITE(*,*) 'everything is OK (no error)'
   CASE (-1)
      WRITE(*,*) 'an illegal interval range was specified (b <= a)'
   CASE (-2)
      WRITE(*,*) 'an illegal smallest step size was specified'
   CASE (-3)
      WRITE(*,*) 'an illegal starting step size was specified'
   CASE (-4)
      WRITE(*,*) 'an illegal global tolerance was specified (tol <= 0)'
   CASE (-5)
      WRITE(*,*) 'an illegal Newton-tolerance was specified'
      WRITE(*,*) '(Newton-tol >= global tol or Newton-tol <= 0)'
   CASE (-6)
     WRITE(*,*) 'an illegal order of method or polynomial was specified'
     WRITE(*,*) '(order of polynomials <= order of method)'
   CASE (-7)
      WRITE(*,*) 'an illegal maximum step size was specified'
   CASE (-8)
     WRITE(*,*) 'an illegal switch point was specified (x not in (a,b))'
   CASE (1)
      WRITE(*,*) 'IVP solver could not obtain specified accuracy', &
                 ' (h = h(global tol) < h_min)'
   CASE (2)
      WRITE(*,*) 'the Newton-iteration needed too many steps', &
                 ' (Newton did not converge)'
   CASE (3)
    WRITE(*,*) 'the global error did not decrease during IDeC iteration'
    WRITE(*,*) '(IDeC did not converge for h to 0)'
   CASE (4)
    WRITE(*,*) 'the Newton-residuals did not decrease during iterations'
    WRITE(*,*) '(Newton-residuals were not decreasing in monotonic way)'
   CASE (5)
      WRITE(*,*) 'the Newton-iteration needed too many steps', &
                 ' (Newton did not converge)'
   CASE (6)
    WRITE(*,*) 'the Newton-residuals did not decrease during iterations'
    WRITE(*,*) '(Newton-residuals were not decreasing in monotonic way)'
   END SELECT
   IF ( (bstop) .AND. (ifail /= 0) ) THEN
      WRITE(*,*) 'FUER ORDALL (dummy)'
      WRITE(*,*) 0.0_dp, 0.0_dp
      STOP
   END IF
   RETURN
END SUBROUTINE error

END PROGRAM SSHT
